/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.util;

import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.ClassNotFoundException;

/**
 * Hashtable (NOT tree) with simple clustering.
 */

public class Hashtable
    extends Dictionary
    implements Cloneable, Serializable
{
    /** The hashtable keys. */
    transient private Object keys[];

    /** The hashtable values of each key.  */
    transient private Object elements[];

    /**
     * The load factor for the hashtable.
     * @serial
     */
    private float loadFactor;

    /**
     * Usable slots, limited by loading-factor.
     * @serial The table is rehashed when its size exceeds this threshold.
     *   (The value of this field is (int)((float)capacity * loadFactor).)
     */
    private int threshold;

    /** Used slots in the table.  */
    transient private int numberOfKeys;

    private static final int DEFAULTCAPACITY = 101;
    private static final float DEFAULTLOADFACTOR = 0.75f;

    /** Magic key value that mark deleted slot,
     * that don't break the collition chain.  */
    private static final Object removedSlot = new Object();

    /**
     * Check if slot is empty.
     */
    private static boolean isEmptySlot(Object key) {
	return (key == null) || (key == removedSlot);
    }

    /* This is what Sun's JDK1.1 "serialver java.util.Hashtable" spits out */
    private static final long serialVersionUID = 1421746759512286392L;

    public Hashtable() {
	this(DEFAULTCAPACITY, DEFAULTLOADFACTOR);
    }

    public Hashtable(int initialCapacity) {
	this(initialCapacity, DEFAULTLOADFACTOR);
    }


    /**
     * Constructs a new Hashtable with a size and an initial loading-factor.
     *
     * @param initialCapacity	the initial number of slots.
     * @param loadFactor the loading-factor between 0.0 and 1.0, it defines
     *     the threshold for rehashing the hashtable into a bigger one.
     */
    public Hashtable(int initialCapacity, float loadFactor) {
	if ((initialCapacity <= 0) ||
	    (loadFactor <= 0.0f) || (loadFactor > 1.0f))
	    throw new IllegalArgumentException();
	this.loadFactor = loadFactor;
	this.keys = new Object[initialCapacity];
	this.elements = new Object[initialCapacity];
	this.numberOfKeys = 0;
	this.threshold = (int)(loadFactor * (float)initialCapacity);
    }


    public int size() {
	return (numberOfKeys);
    }


    public boolean isEmpty() {
	return (numberOfKeys == 0);
    }


    public synchronized Enumeration keys() {
	Vector vector = new Vector(numberOfKeys);

	for (int pos = keys.length; pos-- > 0; ) {
	    if (!isEmptySlot(keys[pos]))
		vector.addElement(keys[pos]);
	}

	return new HashtableEnumeration(vector);
    }


    public synchronized Enumeration elements() {
	Vector vector = new Vector(numberOfKeys);

	for (int pos = elements.length; pos-- > 0; ) {
	    if (!isEmptySlot(keys[pos]))
		vector.addElement(elements[pos]);
	}

	return new HashtableEnumeration(vector);
    }
  

    public synchronized boolean contains(Object value) {
	for (int pos = elements.length; pos-- > 0; ) {
	    // if this slot is empty, elements[pos] == null
	   
	    // Some class could not check for null parameter in there
	    // equals() implementation, polardoc.Names in polardoc 1.0.7
	    // is an exemple.
	    // Prevents NullPointerException.
	    Object elem = elements[pos];
	    if ((elem != null) && value.equals(elem))
		return true;
	}
	return false;
    }

    public synchronized boolean containsKey(Object key) {
	return (get(key) != null);
    }
  

    /**
     * Return the index of the slot matching the key.
     *
     * @param key	the search key
     * @return	the slot index, an empty slot
     *    or -1 iif the table is full and threshold == size().
     */
    private int calculateSlot(Object key) {
	int index = (key.hashCode() & Integer.MAX_VALUE) % keys.length;
	int start_index = index;
	int free_index = -1;
	int step = 8 * keys.length + 7;	// step and keys.length must be prime.

	while (true) {
	    Object match_key = keys[index];
	    if (match_key == null)
		return (free_index == -1) ? index : free_index;
	    else if (match_key == removedSlot) {
		if (free_index == -1)
		    free_index = index;
	    }
	    else if (key.equals(match_key))
		return index;
	    index = (index + step) % keys.length;
	    if (index == start_index)
		return free_index;
	}
    }


    /**
     * Rehashes the content of the table into a bigger table.
     * This method is called automatically when the hashtable's
     * size exceeds the threshold.
     */
    protected synchronized void rehash()
    {
	int newCapacity = keys.length * 2;
	Object oldKeys[] = keys;
	Object oldElements[] = elements;

	keys = new Object[newCapacity];
	elements = new Object[newCapacity];
	threshold = (int)(loadFactor * (float)newCapacity);
	numberOfKeys = 0;

	/* Go through adding all the data to the new data */
	for (int pos = oldKeys.length; pos-- > 0; ) {
	    if (!isEmptySlot(oldKeys[pos]))
		put(oldKeys[pos], oldElements[pos]);
	}
    }


    /**
     * Clear all entries in the hashtable.
     */
    public synchronized void clear() {
	for (int pos = keys.length; pos-- > 0; ) {
	    keys[pos] = null;
	    elements[pos] = null;
	}
	numberOfKeys = 0;
    }
  

    /**
     * Gets the object associated with the specified key in the
     * hashtable.
     * @param key	the specified key, must be not empty.
     * @return the element for the key or null if the key
     *          is not defined in the hash table.
     */
    public synchronized Object get(Object key)
    {
	int index = calculateSlot(key);
	if ((index == -1) || isEmptySlot(keys[index]))
	    return null;
	else
	    return elements[index];
    }
  

    /**
     * Puts the specified element into the hashtable, using the specified
     * key.  The element may be retrieved by doing a get() with the same key.
     * The key and the element cannot be null.
     * @param key	the specified key in the hashtable
     * @param value	the specified element
     * @return the old value of the key, or null if it did not have one.
     */
    public synchronized Object put(Object key, Object value) {
	if (numberOfKeys >= threshold)
	    rehash();

	int index = calculateSlot(key);
	// index == -1 iif numberOfKeys == keys.length,
	// ie keys.length < threshold, ie loadFactor > 1.0
	Object old = elements[index];
	if (isEmptySlot(keys[index])) {
	    keys[index] = key;
	    numberOfKeys++;
	}
	elements[index] = value;
	return old;
    }


    /**
     * Removes the element corresponding to the key.  Does nothing if the
     * key is not present.
     * @param key	the key that needs to be removed
     * @return the element for the key or null if the key
     *          is not defined in the hash table.
     */
    public synchronized Object remove(Object key) {
	int index = calculateSlot(key);
	Object old = null;
	if ((index != -1) && !isEmptySlot(keys[index])) {
	    keys[index] = removedSlot;
	    old = elements[index];
	    elements[index] = null;
	    numberOfKeys--;
	}
	return old;
    }


    /**
     * Creates a shallow copy of this hashtable.  
     * The keys and values themselves are not cloned. 
     * This is a relatively expensive operation. 
     *
     * @return a clone of the hashtable. 
     */
    public synchronized Object clone() {
	Hashtable result;
	try {
	    // Note that we must use super.clone() here instead of a
	    // constructor or else subclasses such as java.util.Properties
	    // will not be cloned properly.
	    result = (Hashtable)super.clone();
	}
	catch (CloneNotSupportedException _) {
	    return (null);
	}
	result.loadFactor = loadFactor;
	result.threshold = threshold;
	result.keys = new Object[keys.length];
	result.elements = new Object[elements.length];
	result.numberOfKeys = 0;

	// copy our entries in new hashtable
	for (int pos = keys.length; pos-- > 0; ) {
	    if (!isEmptySlot(keys[pos]))
		result.put(keys[pos], elements[pos]);
	}
	return ((Object)result);
    }


    /**
     * Read this hashtable from a stream.
     */
    private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
    {
	// read all non-transient fields (aka loadFactor and threshold)
	stream.defaultReadObject();

	// read capacity, and create empty table
	int len = stream.readInt();
	keys = new Object[len];
	elements = new Object[len];
	numberOfKeys = 0;

	// read number of key-value
	int nkeys = stream.readInt();

	// read all key-value
	while (nkeys-- > 0) {
	    Object key = stream.readObject();
	    Object value = stream.readObject();
	    put(key, value);
	}
    }


    /**
     * write this hashtable into a stream.
     * @serialData The capacity of the Hashtable (the length of the bucket
     *   array) is emitted (int), followed by the size of the Hashtable (the
     *   number of key-value mappings), followed by the key (Object) and
     *   value (Object) for each key-value mapping represented by the Hashtable.
     *   The key-value mappings are emitted in no particular order.
     */
    private void writeObject(ObjectOutputStream stream)
	throws IOException
    {
	// write all non-transient fields (aka loadFactor and threshold)
	stream.defaultWriteObject();

	// remember how many slots there were
	stream.writeInt(keys.length);
	
	// write number of key-value
	stream.writeInt(numberOfKeys);

	// write all key-value
	for (int pos = keys.length; pos-- > 0; ) {
	    if (!isEmptySlot(keys[pos])) {
		stream.writeObject(keys[pos]);
		stream.writeObject(elements[pos]);
	    }
	}
    }


    public synchronized String toString() {
	boolean firstTime = true;
	StringBuffer result = new StringBuffer();

	result.append('{');
	for (int pos = keys.length; pos-- > 0; ) {
	    if (!isEmptySlot(keys[pos])) {
		if (!firstTime)
		    result.append(", ");
		else
		    firstTime = false;
		result.append(keys[pos]);
		result.append('=');
		result.append(elements[pos]);
	    }
	}
	result.append('}');

	return (result.toString());
    }
}


class HashtableEnumeration
    implements Enumeration
{
    private Vector vector;
    private int posn = 0;

    public HashtableEnumeration(Vector vector) {
	this.vector = vector;
    }

    public boolean hasMoreElements() {
	return (posn < vector.size());
    }

    public Object nextElement() {
	if (posn >= vector.size())
	    throw new NoSuchElementException();
	return (vector.elementAt(posn++));
    }
}
